/* ==================================================================
   File:        kscore.h
   Author:      Steve Monks
   Description:
	Kernel Streaming playback manager

 	$Id: kscore.h,v 1.8 2006/04/09 19:51:53 Steve Exp $
	$Log: kscore.h,v $
	Revision 1.8  2006/04/09 19:51:53  Steve
	Version 3.6
	Added logarithmic volume control
	Added diagnostic logging
	
	Revision 1.7  2006/03/12 15:57:09  Steve
	Fixed incorrect play reporting issue
	
	Revision 1.6  2006/02/04 21:57:38  Steve
	Added out shutdown of KS when buffer underruns in gapless mode
	tweaks to readme
	
	Revision 1.5  2006/02/04 20:59:44  Steve
	Initial import of readme.html
	
	Added gapless mode (for evaluation purposes)
	
	Revision 1.4  2006/01/29 21:33:57  Steve
	Added an adjustment for the number of output buffers
	Removed buffer size radio buttons and replaced with sliders
	Now deletes and recreates the KSCore object when a new track starts after changing the settings
	
	Revision 1.3  2006/01/08 22:44:37  Steve
	Added status window
	Added increased buffer settings
	Fixed .d00 file problems (returned truthful buffer sizes rather than 8192)
	
	Revision 1.2  2006/01/06 22:26:20  Steve
	Fixed problem with Winamp Music and .mod playback (checking against packet_size
	rather than len inside Write function).
	Removed delay before player thread sleeps when starved of data
	
	Revision 1.1  2006/01/03 21:46:13  Steve
	Initial import of V3.0.
	
	This is a complete rewrite of the plugin and Kernel streaming code.
	The only code retained from the previous 2.x version is the Microsoft
	sample code and the filter enumeration.
	
	
================================================================== */

#ifndef _KSCORE_H
#define _KSCORE_H

class KSCore {
private:

	typedef struct data_packet_s {
	    KSSTREAM_HEADER Header;
	    OVERLAPPED      Signal;
	} DATA_PACKET;


	enum { NUM_PACKETS = 8 };
	enum {
		EV_MSG,		// index of MSG synchronisation object
		EV_KSPACKET	// index of first KS packet event
	};

	typedef enum {
		MODE_STOPPED,
		MODE_STARTED,
		MODE_STOPPING,
		MODE_CNT
	} PLAYER_MODE;

	volatile int paused;									// current pause state, note, thread may still be feeding buffers until thread_paused is set
	volatile int thread_paused;						// set/cleared when thread has synchronised with current pause status
	volatile PLAYER_MODE mode;							// current player mode
	char *input_buffer;									// packets from winamp are copied here temporarily before being transferred to the KS buffers
	int input_buffer_size;								// size of temp buffer
	int num_packets;										// actual number of packets
	volatile int should_be_playing;					// if true, according to WinAmp, we should be playing

	volatile char *read_ptr;							// ptr to data start - modified by player
	volatile char *write_ptr;							// ptr to first free byte - modified by winamp
	volatile int input_buffer_len;						// amount of data currently in temp buffer, set by Winamp after filling the buffer, cleared by player thread after copying the buffer

	
	int volume;												// current volume
	int packet_size;										// size of each packet presented to KS interface
	HWND diagnostic_wnd;
	HWND hwnd;												// parent window (for dialogs etc)
	HINSTANCE hinst;										// parent instance
	DWORD threadid;										// ID of player thread
	HANDLE player_thread;								// handle of player thread
	HANDLE *events;										// array of event handles
	DATA_PACKET packets[NUM_PACKETS];				// packet headers for KS packets
	CKsEnumFilters* the_enumerator;					
	CKsAudRenFilter*	filter;							
	CKsAudRenPin*		pin;								
	char *output_buffer;									// buffer of sample data - split between KS packets
	int cur_packet;										// packet we are currently writing, -1 = not currently writing a packet

	HANDLE sleep_event;
	volatile int dbg_player_sleeps;					// debug parameter to determine state of player thread
	volatile int last_tick;
	CRITICAL_SECTION cs_msg;							
	CRITICAL_SECTION ks_access;						// critical section for talking to KS interface
	volatile int written_data;                   // number of bytes written out by plugin

	class Config *config;								
	WAVEFORMATEXTENSIBLE wfx_cur;						// current output format

	volatile int quit;									// flag used to close down player thread
	volatile KSSTATE state;								// kernel streaming state - not really used

	void FlushInputBuffer(void);						// empties the intermediate buffer
	int CopyFromInput(int pkt, int amt);				// copies amt bytes from temp buffer to packet, updating all relevant pointers and counts
	void CopyMemoryAttenuated(void *dest, void *source, int len);
	int MemCpyWithUnpack(int *out, char *in, int in_len);
	void WakePlayer(void);
	DWORD WINAPI PlayerThread(LPVOID parm);		// the player thread
	static DWORD WINAPI PlayerThreadWrapper(LPVOID parm);
	CKsEnumFilters *OpenEnumerator(void);
	void CloseEnumerator(void);
	CKsAudRenFilter *PickFilter(void);
	int OpenOutput(WAVEFORMATEXTENSIBLE *wfx);	// Initialise the output device
	void CloseOutput(void);								// Release the output device
	void InitPackets(int init_events);				// initialises the packets, if init_events is true, the events are created
	void InitInputBuffer(int size);
	void InitOutputBuffer(int size);

	void DrawDiagnostics(HWND hwnd, HDC hdc);
	VOID CALLBACK DiagnosticRefresh(HWND hwnd, UINT msg, UINT_PTR id, DWORD dwTime);
	static VOID CALLBACK DiagnosticRefreshWrapper(HWND hwnd, UINT msg, UINT_PTR idevent, DWORD dwTime);
	BOOL CALLBACK KSCore::DiagnosticDialog(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
	static BOOL CALLBACK DiagnosticDialogWrapper(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
	void End(int force);
	void EndInternal(int force);														// all End functions eventually end up here
	int StartInternal(int samplerate, int numchannels, int bitspersamp);	// all Start functions eventually end up here
	void ConvertWaveFormatExtensibleToWaveFormat(WAVEFORMATEXTENSIBLE *out, WAVEFORMATEXTENSIBLE *in);
public:
	KSCore(HWND hwnd, HINSTANCE hinst, class Config *config);
	~KSCore();
	int Start(int samplerate, int numchannels, int bitspersamp);
	void End(void);
	int SetPause(int pause_state);
	int WritePacket(char *buf, int len);
	int CanWrite(void);
	int IsPlaying(void);
	int GetWrittenTime(void);
	void SetVolume(int vol);
	CKsEnumFilters* GetEnumerator(void) { return the_enumerator; }

	void OpenDiagnosticWindow(void);
	void CloseDiagnosticWindow(void);
};

#endif


